<?php

/**
 * @file
 * Post update functions for data_pipelines module.
 */

use Drupal\Core\Site\Settings;
use Drupal\data_pipelines\Form\DatasetBatchOperations;
use Drupal\data_pipelines\Index\IndexingRequest;

/**
 * Add machine names to all DataSet entities.
 */
function data_pipelines_post_update_add_machine_name(&$sandbox = NULL) {
  $storage = \Drupal::entityTypeManager()->getStorage('data_pipelines');
  $iteration_size = (int) Settings::get('entity_update_batch_size', 50);
  if (empty($sandbox['progress'])) {
    $sandbox['progress'] = 0;
    $sandbox['max'] = (int) $storage->getQuery()
      ->accessCheck(FALSE)
      ->count()
      ->execute();
  }

  $start = $sandbox['progress'];
  $end = min($sandbox['max'], $start + $iteration_size);

  $ids = $storage->getQuery()
    ->sort('id')
    ->accessCheck(FALSE)
    ->range($start, $end)
    ->execute();
  $datasets = $storage->loadMultiple(array_values($ids));
  $value_taken = function ($id, $machine_name) use ($storage) {
    return (bool) $storage->getQuery()
      ->condition('id', $id, '<>')
      ->condition('machine_name', $machine_name)
      ->range(0, 1)
      ->count()
      ->execute();
  };
  foreach ($datasets as $id => $dataset) {
    $machine_name = preg_replace('@[^a-z0-9_]+@', '_', trim(mb_strtolower($dataset->label())));
    while ($value_taken($id, $machine_name)) {
      $machine_name .= '_' . mt_rand(0, 9);
    }
    $dataset->machine_name->value = $machine_name;
    $dataset->save();
  }
  if ($sandbox['max'] > 0 && $end < $sandbox['max']) {
    $sandbox['progress'] = $end;
    $sandbox['#finished'] = ($end - 1) / $sandbox['max'];
  }
  else {
    $sandbox['#finished'] = 1;
  }
}

/**
 * Recreate all the existing indexes.
 */
function data_pipelines_post_update_delete_elastic_indexes(&$sandbox = NULL) {
  $storage = \Drupal::entityTypeManager()->getStorage('data_pipelines');
  $iteration_size = (int) Settings::get('entity_update_batch_size', 50);
  if (empty($sandbox['progress'])) {
    $sandbox['progress'] = 0;
    $sandbox['max'] = (int) $storage->getQuery()
      ->accessCheck(FALSE)
      ->count()
      ->execute();
  }

  $start = $sandbox['progress'];
  $end = min($sandbox['max'], $start + $iteration_size);

  $ids = $storage->getQuery()
    ->sort('id')
    ->accessCheck(FALSE)
    ->range($start, $end)
    ->execute();
  /** @var \Drupal\data_pipelines\Entity\DatasetInterface[] $datasets */
  $datasets = $storage->loadMultiple(array_values($ids));
  /** @var \Drupal\data_pipelines\Index\DatasetIndexerInterface $indexer */
  $indexer = \Drupal::service('data_pipelines.indexer');
  $set_protected_property = function ($object, $property, $value) {
    $reflection = new ReflectionClass($object);
    $property = $reflection->getProperty($property);
    $property->setAccessible(TRUE);
    $property->setValue($object, $value);
  };
  foreach ($datasets as $dataset) {
    try {
      $indexing_request = IndexingRequest::deleteDataSet($dataset);
      // Update the indexing request as per old logic.
      $set_protected_property($indexing_request, 'machineName', $dataset->uuid());
      $indexer->deleteDataSetFromIndex($indexing_request);
    }
    catch (\Exception $e) {
      // A critical log message is already logged so do nothing.
    }
    try {
      batch_set(DatasetBatchOperations::batchForDataset($dataset));
    }
    catch (\Exception $e) {
      // A critical log message is already logged so do nothing.
    }
  }
  if ($sandbox['max'] > 0 && $end < $sandbox['max']) {
    $sandbox['progress'] = $end;
    $sandbox['#finished'] = ($end - 1) / $sandbox['max'];
  }
  else {
    $sandbox['#finished'] = 1;
  }
}

/**
 * Update all the existing dataset with csv source to have delimiter set to Comma as a default.
 */
function data_pipelines_post_update_set_delimiter(&$sandbox = NULL) {
  $storage = \Drupal::entityTypeManager()->getStorage('data_pipelines');
  $iteration_size = (int) Settings::get('entity_update_batch_size', 50);
  if (empty($sandbox['progress'])) {
    $sandbox['progress'] = 0;
    $sandbox['max'] = (int) $storage->getQuery()
      ->condition('source', 'csv:%', 'LIKE')
      ->accessCheck(FALSE)
      ->count()
      ->execute();
  }
  $start = $sandbox['progress'];
  $end = min($sandbox['max'], $start + $iteration_size);
  $ids = $storage->getQuery()
    ->condition('source', 'csv:%', 'LIKE')
    ->sort('id')
    ->accessCheck(FALSE)
    ->range($start, $end)
    ->execute();
  /** @var \Drupal\data_pipelines\Entity\DatasetInterface[] $datasets */
  $datasets = $storage->loadMultiple(array_values($ids));
  foreach ($datasets as $dataset) {
    $dataset->set('csv_delimiter', 'comma');
    $dataset->save();
  }
  if ($sandbox['max'] > 0 && $end < $sandbox['max']) {
    $sandbox['progress'] = $end;
    $sandbox['#finished'] = ($end - 1) / $sandbox['max'];
  }
  else {
    $sandbox['#finished'] = 1;
  }
}
