<?php

declare(strict_types=1);

namespace Drupal\data_pipelines\Form;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\Queue\RequeueException;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\data_pipelines\Entity\Dataset;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Index\IndexingRequest;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolation;

/**
 * Defines a class for building and running dataset batched operations.
 */
class DatasetBatchOperations implements ContainerInjectionInterface {

  /**
   * Queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Queue manager.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $queueManager;

  /**
   * Constructs a new PreviewSiteBuilder.
   *
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   Queue factory.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $queueManager
   *   Queue manager.
   */
  final public function __construct(QueueFactory $queueFactory, QueueWorkerManagerInterface $queueManager) {
    $this->queueFactory = $queueFactory;
    $this->queueManager = $queueManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('queue'),
      $container->get('plugin.manager.queue_worker')
    );
  }

  /**
   * Factory method.
   *
   * @return \Drupal\data_pipelines\Form\DatasetBatchOperations
   *   New instance.
   */
  protected static function factory(): DatasetBatchOperations {
    $instance = \Drupal::classResolver(self::class);
    assert($instance instanceof DatasetBatchOperations);
    return $instance;
  }

  /**
   * Processes a queue item.
   *
   * @param string $queue_name
   *   Queue name.
   *
   * @return int
   *   Number of remaining items.
   */
  protected function processQueueItem(string $queue_name) : int {
    $queue = $this->queueFactory->get($queue_name);
    $item = $queue->claimItem();
    if (!$item) {
      return 0;
    }
    $queue->createQueue();
    $worker = $this->queueManager->createInstance($queue_name);
    try {
      $worker->processItem($item->data);
      $queue->deleteItem($item);
    }
    catch (RequeueException $e) {
      $queue->releaseItem($item);
    }
    return $queue->numberOfItems();
  }

  /**
   * Gets batch for dataset.
   *
   * @param \Drupal\data_pipelines\Entity\DatasetInterface $dataset
   *   Dataset.
   * @param bool $saving
   *   TRUE if Dataset is being saved FALSE otherwise.
   *
   * @return array
   *   Batch.
   */
  public static function batchForDataset(DatasetInterface $dataset, bool $saving = FALSE): array {
    $batch_builder = new BatchBuilder();
    $batch_builder
      ->setTitle(new TranslatableMarkup('Validating and indexing dataset %name', [
        '%name' => $dataset->label(),
      ]))
      ->setInitMessage(new TranslatableMarkup('Preparing...'))
      ->setProgressive(TRUE)
      ->setFinishCallback([self::class, 'finished'])
      ->setProgressMessage(new TranslatableMarkup('Step @current of @total'))
      ->addOperation([self::class, 'operationValidate'], [$dataset->id()]);
    // After saving the Dataset, it is already queued in
    // \Drupal\data_pipelines\EntityHandlers\DatasetStorage::save() so we can
    // skip this step here.
    if (!$saving) {
      $batch_builder
        ->addOperation([self::class, 'operationQueueItem'], [$dataset->id()]);
    }
    $batch_builder
      ->addOperation([self::class, 'operationIndex'], [$dataset->id()]);
    return $batch_builder->toArray();
  }

  /**
   * Batch callback.
   */
  public static function operationValidate(int $dataset_id, &$context): void {
    $context['results']['dataset_id'] = $dataset_id;
    $dataset = Dataset::load($dataset_id);
    $dataset->resetLogs();
    assert($dataset instanceof DatasetInterface);
    $errors = $dataset->validate();
    if ($errors->count() > 0) {
      foreach ($errors as $validation_error) {
        assert($validation_error instanceof ConstraintViolation);
        $dataset->addLogMessage((string) $validation_error->getMessage(), FALSE);
      }
      $dataset->setPendingValidation();
      $context['message'] = new TranslatableMarkup('Validation complete, validation errors exist for dataset (@dataset).', ['@dataset' => $dataset->label()]);
      return;
    }
    $dataset->setPendingIndex();
    $context['message'] = new TranslatableMarkup('Validation complete.');
  }

  /**
   * Batch callback.
   */
  public static function operationQueueItem(int $dataset_id, &$context): void {
    $dataset = Dataset::load($dataset_id);
    assert($dataset instanceof DatasetInterface);
    self::factory()->queue($dataset);
    $context['message'] = new TranslatableMarkup('Data is successfully queued for indexing.');
  }

  /**
   * Queue a dataset for indexing.
   *
   * @param \Drupal\data_pipelines\Entity\DatasetInterface $dataset
   *   Dataset.
   *
   * @return int
   *   Number items.
   */
  protected function queue(DatasetInterface $dataset): int {
    $queue = $this->queueFactory->get('data_pipelines_index:' . $dataset->id());
    // Clear any previous operations.
    $queue->deleteQueue();
    $queue->createItem(IndexingRequest::forDataSet($dataset));
    return $queue->numberOfItems();
  }

  /**
   * Index a dataset.
   *
   * @param \Drupal\data_pipelines\Entity\DatasetInterface $dataset
   *   Dataset.
   *
   * @return int
   *   Number remaining.
   */
  protected function index(DatasetInterface $dataset): int {
    return $this->processQueueItem('data_pipelines_index:' . $dataset->id());
  }

  /**
   * Batch callback.
   */
  public static function operationIndex(int $dataset_id, &$context): void {
    $dataset = Dataset::load($dataset_id);
    assert($dataset instanceof DatasetInterface);
    if ($dataset->isPendingIndex()) {
      $remaining = self::factory()->index($dataset);
      if ($remaining === 0) {
        $dataset->setIndexingComplete();
        $context['message'] = new TranslatableMarkup('Indexed dataset.');
        return;
      }
    }
    $context['message'] = new TranslatableMarkup('Skipped indexing dataset, validation errors exist.');
  }

  /**
   * Batch finished callback.
   */
  public static function finished(bool $success, array $results) {
    if (!empty($results['dataset_id'])) {
      $dataset = Dataset::load($results['dataset_id']);
      assert($dataset instanceof DatasetInterface);
      if ($dataset->isIndexed()) {
        $message = new TranslatableMarkup('Dataset %name was successfully indexed.', ['%name' => $dataset->label()]);
        static::log($message);
        return;
      }
      $message = new TranslatableMarkup('Dataset %name was unable to be indexed. Please check <a href=":url">the logs</a>.', [
        '%name' => $dataset->label(),
        ':url' => $dataset->toUrl()->toString(),
      ]);
      static::log($message, TRUE);
      return;
    }
    $message = new TranslatableMarkup('The validation and indexing operation did not complete.');
    static::log($message, TRUE);
  }

  /**
   * A method to facilitate logging.
   *
   * @param \Drupal\Component\Render\MarkupInterface $message
   *   The message.
   * @param bool $error
   *   A boolean indicating whether the log is an error.
   */
  protected static function log(MarkupInterface $message, bool $error = FALSE): void {
    \Drupal::logger('data_pipelines')->{$error ? 'warning' : 'info'}($message);
    \Drupal::messenger()->{$error ? 'addError' : 'addMessage'}($message);
  }

}
