<?php

declare(strict_types = 1);

namespace Drupal\data_pipelines\Form;

use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityConstraintViolationListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\data_pipelines\Entity\Dataset;
use Drupal\data_pipelines\Entity\DatasetInterface;

/**
 * Provides a form for adding/editing datasets.
 *
 * @codeCoverageIgnore
 * @see \Drupal\Tests\data_pipelines\Functional\DatasetUiTest
 */
class DatasetForm extends ContentEntityForm {

  /**
   * Flag to track if validation is occurring.
   *
   * @var bool
   */
  protected $isValidating = FALSE;

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $dataset = $this->entity;
    assert($dataset instanceof DatasetInterface);

    if ($this->operation === 'edit') {
      $form['#title'] = $this->t('Edit dataset %label', ['%label' => $dataset->label()]);
    }

    // @todo remove these changes
    //   https://www.drupal.org/project/drupal/issues/2685749 is fixed.
    $form['machine_name'] = [
      '#type' => 'machine_name',
      '#title' => $this->t('Machine Name'),
      '#maxlength' => 128,
      '#weight' => -4,
      '#default_value' => $dataset->getMachineName(),
      '#disabled' => !$dataset->isNew(),
      '#machine_name' => [
        'exists' => [Dataset::class, 'loadByMachineName'],
      ],
      '#element_validate' => [],
    ];

    return parent::form($form, $form_state);
  }

  /**
   * {@inheritdoc}
   *
   * @todo remove this method once
   *   https://www.drupal.org/project/drupal/issues/2685749 is fixed.
   */
  protected function getEditedFieldNames(FormStateInterface $form_state) {
    return array_merge([
      'machine_name',
    ], parent::getEditedFieldNames($form_state));
  }

  /**
   * {@inheritdoc}
   *
   * @todo remove this method once
   *   https://www.drupal.org/project/drupal/issues/2685749 is fixed.
   */
  protected function flagViolations(EntityConstraintViolationListInterface $violations, array $form, FormStateInterface $form_state) {
    // Manually flag violations of fields not handled by the form display. This
    // is necessary as entity form displays only flag violations for fields
    // contained in the display.
    $field_names = [
      'machine_name',
    ];
    foreach ($violations->getByFields($field_names) as $violation) {
      [$field_name] = explode('.', $violation->getPropertyPath(), 2);
      $form_state->setErrorByName($field_name, $violation->getMessage());
    }
    parent::flagViolations($violations, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $entity = parent::buildEntity($form, $form_state);
    if ($this->isValidating) {
      // We set the syncing flag to TRUE to avoid running validation of the data
      // from the form operation. As if the dataset is large, we may be unable
      // to complete validation in a single request.
      $entity->setSyncing(TRUE);
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $this->isValidating = TRUE;
    $entity = parent::validateForm($form, $form_state);
    $this->isValidating = FALSE;
    $entity->setSyncing(FALSE);
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $dataset = $this->entity;
    assert($dataset instanceof DatasetInterface);

    $result = $dataset->save();

    $edit_link = $dataset->toLink($this->t('Edit'), 'edit-form')->toString();
    $view_link = $dataset->toLink()->toString();
    switch ($result) {
      case SAVED_NEW:
        $this->messenger()->addStatus($this->t('Created dataset %dataset.', [
          '%dataset' => $view_link,
        ]));
        $this->logger('data_pipelines')->notice('Created dataset %dataset.', [
          '%dataset' => $dataset->label(),
          'link' => $edit_link,
        ]);
        break;

      case SAVED_UPDATED:
        $this->messenger()->addStatus($this->t('Saved dataset %dataset.', [
          '%dataset' => $view_link,
        ]));
        $this->logger('data_pipelines')->notice('Saved dataset %dataset.', [
          '%dataset' => $dataset->label(),
          'link' => $edit_link,
        ]);
        break;
    }
    batch_set(DatasetBatchOperations::batchForDataset($dataset, TRUE));
    $form_state->setRedirectUrl($dataset->toUrl('collection'));
    return $result;
  }

}
