<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines\Kernel\Index;

use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\data_pipelines\DatasetData;
use Drupal\data_pipelines\Index\DatasetElasticsearchIndexer;
use Drupal\data_pipelines\Index\DatasetNotRemovedFromIndexException;
use Drupal\data_pipelines\Index\IndexingRequest;
use Drupal\Tests\data_pipelines\Kernel\DatasetKernelTestBase;
use Elasticsearch\Client as BaseClient;
use Elasticsearch\Namespaces\IndicesNamespace;
use nodespark\DESConnector\Client;
use Prophecy\Argument;

/**
 * Defines a class for testing elastic indexer.
 *
 * @group data_pipelines
 *
 * @covers \Drupal\data_pipelines\Index\DatasetElasticsearchIndexer
 * @covers \Drupal\data_pipelines\Index\DatasetNotRemovedFromIndexException
 */
class DatasetElasticsearchIndexerTest extends DatasetKernelTestBase {

  /**
   * Module config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();
    $this->config = \Drupal::config('data_pipelines.settings');
  }

  /**
   * Tests index creation.
   */
  public function testIndexCreation() {
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(FALSE);
    $dataset = $this->createTestDataset();

    \Drupal::configFactory()->getEditable('data_pipelines.settings')->set('index_name_prefix', 'foo_')->save();
    $indices->create([
      'index' => 'foo_' . $dataset->getMachineName(),
      'body' => [
        'settings' => [
          'number_of_shards' => DatasetElasticsearchIndexer::INDEX_SHARDS,
          'number_of_replicas' => DatasetElasticsearchIndexer::INDEX_REPLICAS,
          'codec' => DatasetElasticsearchIndexer::INDEX_CODEC,
        ],
      ],
    ])->shouldBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());
    $client->CheckResponseAck(Argument::any())->willReturn(TRUE);

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), \Drupal::config('data_pipelines.settings'));
    $indexer->indexDataSet($dataset);
  }

  /**
   * Tests index creation with mapping.
   */
  public function testIndexCreationWithMapping() {
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(FALSE);
    $dataset = $this->createTestDataset(['pipeline' => 'test_pipeline_2']);

    \Drupal::configFactory()->getEditable('data_pipelines.settings')->set('index_name_prefix', 'foo_')->save();
    $indices->create([
      'index' => 'foo_' . $dataset->getMachineName(),
      'body' => [
        'settings' => [
          'number_of_shards' => DatasetElasticsearchIndexer::INDEX_SHARDS,
          'number_of_replicas' => DatasetElasticsearchIndexer::INDEX_REPLICAS,
          'codec' => DatasetElasticsearchIndexer::INDEX_CODEC,
        ],
        'mappings' => [
          'properties' => [
            'firstname' => [
              'type' => 'text',
              'index' => FALSE,
            ],
            'lastname' => [
              'type' => 'text',
              'index' => FALSE,
            ],
            'full_name' => [
              'type' => 'text',
            ],
          ],
        ],
      ],
    ])->shouldBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());
    $client->CheckResponseAck(Argument::any())->willReturn(TRUE);

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), \Drupal::config('data_pipelines.settings'));
    $indexer->indexDataSet($dataset);
  }

  /**
   * Tests index not created if exists.
   */
  public function testIndexNotCreatedIfExists() {
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(TRUE);
    $indices->create(Argument::any())->shouldNotBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());
    $client->CheckResponseAck(Argument::any())->willReturn(TRUE);

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), $this->config);
    $dataset = $this->createTestDataset();
    $indexer->indexDataSet($dataset);
  }

  /**
   * Tests index creation failure.
   */
  public function testIndexCreationFailure() {
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willThrow(new \Exception('Barf'));
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), $this->config);
    $dataset = $this->createTestDataset();
    $this->assertFalse($indexer->indexDataSet($dataset));
  }

  /**
   * Tests index creation failure.
   */
  public function testIndexCreationAckFailure() {
    $indices = $this->prophesize(IndicesNamespace::class);
    $indices->exists(Argument::any())->willReturn(FALSE);
    $indices->create(Argument::any())->shouldBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());
    $client->CheckResponseAck(Argument::any())->willReturn(FALSE);

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), $this->config);
    $dataset = $this->createTestDataset();
    $this->assertFalse($indexer->indexDataSet($dataset));
  }

  /**
   * Tests indexing.
   */
  public function testIndexing() {
    \Drupal::configFactory()->getEditable('data_pipelines.settings')->set('index_name_prefix', 'foo_')->save();
    $indices = $this->prophesize(IndicesNamespace::class);
    $dataset = $this->createTestDataset();
    $indices->exists(Argument::any())->willReturn(TRUE);
    $client = $this->prophesize(BaseClient::class);
    $client->indices()->willReturn($indices->reveal());
    foreach ($dataset->data() as $delta => $data) {
      assert($data instanceof DatasetData);
      $expected = [
        'index' => 'foo_' . $dataset->getMachineName(),
        'id' => $dataset->getMachineName() . ':' . $delta,
        'body' => $data->getArrayCopy(),
      ];
      $client->index($expected)->shouldBeCalled();
    }

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), \Drupal::config('data_pipelines.settings'));
    $indexer->indexDataSet($dataset);
  }

  /**
   * Tests deletion.
   */
  public function testDeleting() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesize(IndicesNamespace::class);
    $client = $this->prophesize(BaseClient::class);
    $client->indices()->willReturn($indices->reveal());
    $expected = [
      'index' => $dataset->getMachineName(),
    ];
    $indices->delete($expected)->shouldBeCalled();

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), $this->config);
    $indexer->deleteDataSetFromIndex(IndexingRequest::deleteDataSet($dataset));
  }

  /**
   * Tests deletion failure.
   */
  public function testDeletingException() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesize(IndicesNamespace::class);
    $client = $this->prophesize(BaseClient::class);
    $client->indices()->willReturn($indices->reveal());
    $expected = [
      'index' => $dataset->getMachineName(),
    ];
    $indices->delete($expected)->willThrow(new \Exception('barf'));

    $indexer = new DatasetElasticsearchIndexer($client->reveal(), $this->prophesize(LoggerChannelInterface::class)->reveal(), $this->config);
    $this->expectException(DatasetNotRemovedFromIndexException::class);
    $this->expectExceptionMessage(sprintf('Could not remove dataset %s from the index, error was barf', $dataset->label()));
    $indexer->deleteDataSetFromIndex(IndexingRequest::deleteDataSet($dataset));
  }

}
